
struct coeffs_struct
  x::Vector{Float64}
  p::Vector{Float64}
  q::Vector{Float64}
  r::Vector{Float64}

  function coeffs_struct(np::Int64)
    return new(zeros(np), zeros(np), zeros(np), zeros(np))
  end
end

struct ddcoeffs_struct
  x::Vector{Float64}

  p::Vector{Float64}
  q::Vector{Float64}
  r::Vector{Float64}

  dp::AbstractArray
  dq::AbstractArray
  dr::AbstractArray

  ddp::AbstractArray
  ddq::AbstractArray
  ddr::AbstractArray

  function ddcoeffs_struct(np::Int64)
    return new(zeros(np),
               zeros(np), zeros(np), zeros(np),
               zeros(np,2), zeros(np,2), zeros(np,2),
               zeros(np,2,2), zeros(np,2,2), zeros(np,2,2))
  end
end


struct BalloonParams
  k_w::Int64
  l_geom_input::Bool
  l_tokamak_input::Bool
  kth::Int64

  r0::Float64
  amin::Float64
  beta0::Float64
  b0_v::Float64
end


abstract type BallooningCoefInterp end

# Chebyshev x Fourier interpolation of the ballooning eigenvalue problem coefficients in Clebsch x Toroidal coordinates
struct FCInterp <: BallooningCoefInterp
  c2     ::AbstractArray;
  κ      ::AbstractArray;
  bsuppar::AbstractArray;
  nfp::Integer;
  iota::Number;
end

# spline interpolation of the ballooning eigenvalue problem coefficients in Clebsch x Toroidal coordinates
struct SInterp <: BallooningCoefInterp
  c2     ::AbstractArray;
  κ      ::AbstractArray;
  bsuppar::AbstractArray;

  nfp::Integer;
  iota::Number;
end


"""
    cobra(vmec_data, vmec, surfs, nwells, θs, ζs, mode=1, geom_input=true, tokamak_input=false, lscreen=true)

Calculates ballooning equilibrium based off of the CobraVmec code. Originally converted from fortran to Julia by Sanket Patil, rewritten by Aaron Bader

# Arguments
 - `vmec::VMEC.Vmec`: A Vmec struct
 - `slist::Vector{Int64}`: list of s values
 - `nwells::Int64`: Number of helical wells to include
# Optional inputs
 - `θs::Vector{Float64}`: Vector of starting θ values (in radians)
 - `ζs::Vector{Float64}`: Vector of starting ζ values (in radians normalized to field periods)
 - `mode::Int64`: mode number, 1 is the most unstable mode
 - `geom_input::Bool`: Flag, default true
 - `tokamak_input::Bool`: Flag, default false, not sure we'll ever use this
 - `lscreen::Bool`: Flag, output to screen, default true

"""
function cobra(vmec::Vmec, slist::AbstractVector, nwells::Int64;
               θs::AbstractVector = [0.0, π/2, π, 2*π/3, 2*π],
               ζs::AbstractVector = [0.0, π/8, π/4, 2*π/8, π/2],
               mode::Int64=1, geom_input::Bool= true, tokamak_input::Bool=false,
               lscreen::Bool=false, print_progress::Bool=false,
               interp::Bool=true, reltol::Float64=1e-4, np::Int64=793,
               spline::Bool=true)
  #=
  _____________________________________________________________________________________
  |                                                                                   |
  |              COBRA (COde for Ballooning Rapid Analysis)                           |
  |                     ==       =          =     =                                   |
  |                                                                                   |
  |      VERSION: julia rewrite of                                                    |
  |               4.1  (FORTRAN 90: fixed FORMAT; standard I/O; VMEC coordinates)     |
  |                    (Accepts ASYMMETRIC input from VMEC via LASYM)                 |
  |                    (Accepts initial line location as geometrical or (ALHPA,ANGLE) |
  |                    (Good for LRFP=T or F)                                         |
  |      Last update: 03/27/13                                                        |
  |                                                                                   |
  |      AUTHORS:   R. Sanchez                                                        |
  |             Universidad Carlos III de Madrid, Leganes 28911, SPAIN.               |
  |                 S.P. Hirshman                                                     |
  |             Oak Ridge National Laboratory, Oak Ridge, TN 37831-9701, USA          |
  |      JULIA REWRITE: S. Patil, UW-Madison                                          |                                                                             |
  |      REFERENCES:                                                                  |
  |                                                                                   |
  |        1."COBRA: an optimized code for fast analysis of ideal ballooning         |
  |        stability of 3-D magnetic equilibria", R. Sanchez, S.P. Hirshman, J.C.     |
  |        Whitson and A.S. Ware, submitted to Journal of Computational Physics (1999)|
  |                                                                                   |
  |        2."Improved magnetic coordinate representation for ideal ballooning        |
  |        stability calculations with the COBRA code", R. Sanchez, S.P. Hirshman     |
  |        H.V. Wong, to be submitted to Journal of Computational physics (2000)      |
  |                                                                                   |
  |      DISCLAIMER:  This code is under development by R.Sanchez at the Departamento |
  |        de Fisica, Universidad Carlos III de Madrid, SPAIN. As a BETA version, the |
  |        code is supplied on "as it is" basis, and the non-existence of "bugs" of   |
  |        ANY kind is NOT guaranteed. Any problem or comment should be reported to   |
  |        R. Sanchez at raul.sanchez@uc3m.es                                         |
  _____________________________________________________________________________________
  =#
  #this will still crash if surfaces are too small, can determine minimum value
  #but better to allow extrapolation
  slist = slist[slist .> 0.001]
  slist = slist[slist .< 0.999]
  ns = length(slist)
  vmec_data = VmecData(vmec)

  # Add here:  date and time if wanted
  #vmec, vmec_data = VMEC.readVmecWout(extension)
  (r0, amin, beta0, b0_v) = order_input(vmec_data);

  params = BalloonParams(nwells, geom_input, tokamak_input, mode, r0, amin, beta0, b0_v)
  # println("wout file read successfully!")


  # Solve ballooning equation

  balloonGrowth = zeros(length(slist), length(θs), length(ζs))
  mercier = zeros(length(slist))
  pres = zeros(length(slist))
  xmax = zeros(length(slist), length(θs), length(ζs))
  tsym = zeros(Int8, length(slist), length(θs), length(ζs))
  μ0 = 4*π*1.0E-7
  for (i_s,s) in enumerate(slist)
    vmecSurface = VmecSurface(s, vmec);
    pres[i_s] = vmecSurface.pres[1]*μ0
    mercier[i_s] = vmecSurface.DMerc[1]

    if print_progress
      println("i_s = $(i_s)/$(length(slist)), ι = $(vmecSurface.iota[1])"); flush(stdout);
    end


    bci = FCInterp([],[],[],0,0)
    if interp
      bci = BallooningCoefInterp(vmecSurface, params; reltol, spline)
    end

    for (iθ, init_theta) in enumerate(θs)             # Loop over all poloidal, toroidal initial pairs
      for (iζ, init_zeta) in enumerate(ζs)
        #println("at: ",s," ",init_theta," ",init_zeta)

        # the following to be used as args in get_ballooning_grate
        init_zeta = mod(init_zeta, 2*π)/vmec.nfp;
        init_theta = mod(init_theta, 2*π);

        lfail_balloon = false

        # get growth rates on desired surfaces

        bg, xm = get_ballooning_grate(vmecSurface, init_zeta, init_theta, params, bci, interp, np);
        balloonGrowth[i_s, iθ, iζ] = bg
        xmax[i_s, iθ, iζ] = xm
      end
    end
  end

  if lscreen
    for (iθ, θ) in enumerate(θs)
      for (iζ, ζ) in enumerate(ζs)
        println("-"^125)
        if params.l_geom_input
          @printf("   NS      FLUX-s        ZT_0        TH_0       GR. RATE    IT  POINTS        XMAX      OK?     SYMM        PRES         MERC\n")
        elseif params.l_tokamak_input
          @printf("   NS      FLUX-s       ZETAK    ALPHA_ST       GR. RATE    IT  POINTS        XMAX      OK?     SYMM        PRES         MERC\n")
        else
          @printf("   NS      FLUX-s      THETAK   ALPHA_TOK       GR. RATE    IT  POINTS        XMAX      OK?     SYMM        PRES         MERC\n")
        end
        println("-"^125)
        ζp = ζ*180/π / vmec.nfp
        θp = θ*180/π

        for (i_s, s) in enumerate(slist)
          surf = Int(round((s * (vmec.ns - 1)) + 1))
          j = nRichardson[i_s, iθ, iζ]
          bg = balloonGrowth[i_s, iθ, iζ]
          xm = xmax[i_s, iθ, iζ]
          if bg == NaN
            l_success = "F"
          else
            l_success = "T"
          end
          if params.l_geom_input
            @printf("%5i    %8.2e    %8.2e    %8.2e    %11.4e    %2i    %4i    %8.2e    %5s    %8.2e    %8.2e\n",
                    surf, s, ζp, θp, bg, j, np, xm, l_success, pres[i_s], mercier[i_s])
          elseif params.l_tokamak_input
            @printf("%5i    %8.2e    %8.2e    %8.2e    %11.4e    %2i    %4i    %8.2e    %5s    %8.2e    %8.2e\n",
                    surf, s, θp, ζp, bg, j, np, xm, l_success, pres[i_s], mercier[i_s])
          else
            @printf("%5i    %8.2e    %8.2e    %8.2e    %11.4e    %2i    %4i    %8.2e    %5s    %8.2e    %8.2e\n",
                    surf, s, ζp, θp, bg, j, np, xm, l_success, pres[i_s], mercier[i_s])
          end
        end

        if params.l_geom_input
            @printf("ZETA0 = %10.3e THETA0 = %10.3e\n", ζ, θ);
        elseif params.l_tokamak_input
            @printf("ALPHA_TOK = %10.3e THETAK = %10.3e\n", ζ, θ);
        else
            @printf("ALPHA_ST = %10.3e ZETAK = %10.3e\n", θ, ζ);
          end
        end
           @printf("____________________________________________________\n");


    end # zeta loop
  end # theta loop

  return balloonGrowth, mercier, pres
end


"""
    cobra_opt(vmec_data, slist, nwells; θs, ζs, mode, geom_input, tokamak_input,
                   lscreen, print_progress, interp, reltol)

Calculates locations of maximally unstable eigenvalues

# Arguments
 - `vmec::VMEC.Vmec`: A Vmec struct
 - `slist::AbstractVector`: list of s values
 - `nwells::Integer`: Number of helical wells to include
# Optional inputs
 - `θs::AbstractVector`: Vector of starting θ values (in radians)
 - `ζs::AbstractVector`: Vector of starting ζ values (in radians normalized to field periods)
 - `reltol::Number`: Relative tolerance input for coefficient interpolation
 - `np::Integer`: Number of points used in eigenvalue calculation
 - `print_progress::Bool`: Print every time a new s-surface is used (adds excitement)
"""
function cobra_opt(vmec::Vmec, slist::AbstractVector, nwells::Integer;
                   θs::AbstractVector = (0:3).*(2π/4),
                   ζs::AbstractVector = (0:3).*(2π/4),
                   print_progress::Bool=false, reltol::Number=1e-4,
                   np::Integer = 793, spline=true)

  # MAX: This should probably throw an error / warning instead of removing points
  slist = slist[slist .> 0.001]
  slist = slist[slist .< 0.999]

  ns = length(slist)
  nθ = length(θs)
  nζ = length(ζs)

  vmec_data = VmecData(vmec)
  (r0, amin, beta0, b0_v) = order_input(vmec_data);

  # I'm pretty sure these flags are deprecated
  geom_input = true;
  tokamak_input = false;
  mode = 1;
  params = BalloonParams(nwells, geom_input, tokamak_input, mode, r0, amin, beta0, b0_v)

  # Solve ballooning optimization
  balloonGrowth = zeros(ns, nθ, nζ)
  balloonGrowthLoc = zeros(2, ns, nθ, nζ);

  μ0 = 4*π*1.0E-7
  for (i_s,s) in enumerate(slist)

    # Interpolate the surface for coefficients
    vmecSurface = VmecSurface(s, vmec);
    bci = BallooningCoefInterp(vmecSurface, params; reltol, spline)
    ι = get_iota(bci);

    if print_progress
      println("i_s = $(i_s)/$(length(slist)), ι = $(ι)"); flush(stdout);
    end

    # Get function that returns eigenvalue matrices for eig_opt
    xmax = π*(2*params.k_w - 1) ./ (2*vmecSurface.nfp*vmecSurface.iota[1])
    np  = Int(6 * floor(np / 6) + 1);      # number of points always odd and multiple of 3
    xmin = -xmax;
    h   = 2*(xmax/(np - 1));
    KMfun = getmatrix_function(xmin, h, np, bci)


    # Loop over all poloidal, toroidal initial pairs
    for (iθ, init_theta) in enumerate(θs), (iζ, init_zeta) in enumerate(ζs)
      loc, λ = eig_opt(KMfun, [init_theta, init_zeta], vmec.nfp)
      balloonGrowth[i_s, iθ, iζ] = λ;
      balloonGrowthLoc[:, i_s, iθ, iζ] = loc
    end
  end

  return balloonGrowth, balloonGrowthLoc
end
